//this file is part of eMule
//Copyright (C)2002-2007 Merkur ( strEmail.Format("%s@%s", "devteam", "emule-project.net") / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#include "stdafx.h"
#include "emule.h"
#include "UploadQueue.h"
#include "Packets.h"
#include "KnownFile.h"
#include "ListenSocket.h"
#include "Exceptions.h"
#include "Scheduler.h"
#include "PerfLog.h"
#include "UploadBandwidthThrottler.h"
#include "ClientList.h"
#include "LastCommonRouteFinder.h"
#include "DownloadQueue.h"
#include "FriendList.h"
#include "Statistics.h"
#include "MMServer.h"
#include "OtherFunctions.h"
#include "UpDownClient.h"
#include "SharedFileList.h"
#include "KnownFileList.h"
#include "Sockets.h"
#include "ClientCredits.h"
#include "Server.h"
#include "ServerList.h"
#include "WebServer.h"
#include "emuledlg.h"
#include "ServerWnd.h"
#include "TransferWnd.h"
#include "SearchDlg.h"
#include "StatisticsDlg.h"
#include "Kademlia/Kademlia/Kademlia.h"
#include "Kademlia/Kademlia/Prefs.h"
#include "Log.h"
#include "collection.h"
#include <Math.h> //>>> Spike2::L2HAC [enkeyDEV]
#include "Sion/Argos/Argos.h" //dlarge Argos
//>>> WiZaRd::PNCheck
#include "./EMF/eMuleFuture.h" 
#include "TaskbarNotifier.h"
//<<< WiZaRd::PNCheck

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//>>> WiZaRd::Automatic Firewalled Retries
static	uint8	m_iFirewalledRetries = 0;
static	uint32	m_iFirewalledTimer = NULL;	
//<<< WiZaRd::Automatic Firewalled Retries

static uint32 counter, sec, statsave;
static UINT _uSaveStatistics = 0;
// -khaos--+++> Added iupdateconnstats...
static uint32 igraph, istats, iupdateconnstats;
// <-----khaos-


CUploadQueue::CUploadQueue()
{
	m_uiPaybackCount = 0; //>>> WiZaRd::Payback First
	m_iFirewalledTimer = ::GetTickCount()+MIN2MS(2); //>>> WiZaRd::Automatic Firewalled Retries
//>>> WiZaRd::PNCheck
	VERIFY( (h_PNTimer = SetTimer(0,0,MIN2MS(15),PNTimer)) != NULL ); 
	if (thePrefs.GetVerbose() && !h_PNTimer)
		AddDebugLogLine(true,_T("Failed to create 'private message' timer - %s"),GetErrorMessage(GetLastError()));
//<<< WiZaRd::PNCheck
	VERIFY( (h_timer = SetTimer(0,0,100,UploadTimer)) != NULL );
	if (thePrefs.GetVerbose() && !h_timer)
		AddDebugLogLine(true,_T("Failed to create 'upload queue' timer - %s"),GetErrorMessage(GetLastError()));
	m_uiFellBelowSlotsNeeded = NULL; //>>> Add2Up Timer
	datarate = 0;
	counter=0;
	successfullupcount = 0;
	failedupcount = 0;
	totaluploadtime = 0;
	m_nLastStartUpload = 0;
	statsave=0;
	// -khaos--+++>
	iupdateconnstats=0;
	// <-----khaos-
	m_dwRemovedClientByScore = ::GetTickCount();
    m_iHighestNumberOfFullyActivatedSlotsSinceLastCall = 0;
    m_MaxActiveClients = 0;
    m_MaxActiveClientsShortTime = 0;

    m_lastCalculatedDataRateTick = 0;
    m_avarage_dr_sum = 0;
    friendDatarate = 0;

    m_dwLastResortedUploadSlots = 0;

//dlarge Smart Upload Control v2 
	AvgRespondTime[0]=500;
	AvgRespondTime[1]=thePrefs.GetSUCPitch()/2;//Changed by Yun.SF3 (this is too much divided, original is 3)
	MaxVUR=512*(thePrefs.GetMaxUpload()+thePrefs.GetMinUpload()); //When we start with SUC take the middle range for upload
//dlarge end
}

//>>> WiZaRd::Payback First
INT64	GetUpDownDiff(const CUpDownClient* client)
{
	if(client->Credits() == NULL)
		return _I64_MIN;
	return  client->Credits()->GetDownloadedTotal()-client->Credits()->GetUploadedTotal();
}

BOOL IsPayback(const CUpDownClient* client)
{	
	return (!client->GetRequestFile() || client->GetDownloadState() > DS_ONQUEUE)
		&& GetUpDownDiff(client) > (SESSIONMAXTRANS/2);
}

BOOL SecondIsBetterClient(const bool bPaybackCheck, const CUpDownClient* first, const CUpDownClient* second, const UINT firstscore, const UINT secondscore)
{
	if(second == NULL)
		return FALSE;
	if(first == NULL)
		return TRUE;

	if(bPaybackCheck)
	{
		const BOOL bPay1 = IsPayback(first);
		const BOOL bPay2 = IsPayback(second);
		if(bPay1 && !bPay2)
			return FALSE;	//first is better
		if(!bPay1 && bPay2)
			return TRUE;	//second is better
		else if(bPay1 && bPay2) 
			return GetUpDownDiff(second) > GetUpDownDiff(first); //second or first?
		//flow over
	}

	return (secondscore > firstscore);
}
//<<< WiZaRd::Payback First

/**
 * Find the highest ranking client in the waiting queue, and return it.
 *
 * Low id client are ranked as lowest possible, unless they are currently connected.
 * A low id client that is not connected, but would have been ranked highest if it
 * had been connected, gets a flag set. This flag means that the client should be
 * allowed to get an upload slot immediately once it connects.
 *
 * @return address of the highest ranking client.
 */
CUpDownClient* CUploadQueue::FindBestClientInQueue()
{
//>>> WiZaRd::Optimization
	//while I'm at it I cleaned this up a bit...
	//just to mention it... the official client has a bug here
	//we will search for a client with score > 0... now IF the extreme rare case
	//happens that we have only GPL breakers in queue then NOONE will ever be added...
	uint32	bestscore = 0;
	uint32  bestlowscore = 0;
	CUpDownClient* newclient = NULL;
	CUpDownClient* lowclient = NULL;
	
//>>> WiZaRd::PowerShare
	CUpDownClient* friendclient = NULL;
	uint32	bestscorePS = 0;
	uint32  bestlowscorePS = 0;
	CUpDownClient* newclientPS = NULL;
	CUpDownClient* lowclientPS = NULL;
//<<< WiZaRd::PowerShare

//>>> WiZaRd::Payback First
	bool bPaybackCheck = false; 
	switch(thePrefs.GetPaybackFirst())
	{
		default:
		case 0: //off
			break;

		case 1: //on - one slot
			bPaybackCheck = m_lPaybackList.IsEmpty() ? true : false;
			break;

		case 2: //on - once every X slots
			bPaybackCheck = m_uiPaybackCount == 5;
			break;

		case 3: //auto - TODO?
			bPaybackCheck = m_lPaybackList.IsEmpty()
				|| m_uiPaybackCount == 5;
			break;
	}
//<<< WiZaRd::Payback First

	for (POSITION pos1 = waitinglist.GetHeadPosition(); pos1 != NULL;)
	{
		POSITION pos2 = pos1;
		CUpDownClient* cur_client =	waitinglist.GetNext(pos1);
		CKnownFile* file = theApp.sharedfiles->GetFileByID(cur_client->GetUploadFileID());
		//While we are going through this list.. Lets check if a client appears to have left the network..
		ASSERT ( cur_client->GetLastUpRequest() );
		if ((::GetTickCount() - cur_client->GetLastUpRequest() > MAX_PURGEQUEUETIME) || !theApp.sharedfiles->GetFileByID(cur_client->GetUploadFileID()) )
		{
			//This client has either not been seen in a long time, or we no longer share the file he wanted anymore..
			cur_client->ClearWaitStartTime();
                        cur_client->m_bUpIsProblematic = false; //dlarge UploadingProblemClient
			RemoveFromWaitingQueue(pos2,true);
			continue;
		}

//>>> WiZaRd::PowerShare
		if(cur_client->GetFriendSlot())
			friendclient = cur_client;
		if(friendclient)
			continue;
//<<< WiZaRd::PowerShare
		// finished clearing
		const uint32 cur_score = cur_client->GetScore(false);

//>>> WiZaRd::PowerShare
		if(file->IsPowerShared())
		{
//>>> WiZaRd::Payback First
			if(SecondIsBetterClient(bPaybackCheck, newclientPS, cur_client, bestscorePS, cur_score))
			//if (cur_score > bestscorePS)
//<<< WiZaRd::Payback First
			{
				// cur_client is more worthy than current best client that is ready to go (connected).
				if(!cur_client->HasLowID() 
                                && (!cur_client->m_bUpIsProblematic) //dlarge UploadingProblemClient
                                || (cur_client->socket && cur_client->socket->IsConnected())) 
				{
					// this client is a HighID or a lowID client that is ready to go (connected)
					// and it is more worthy
					bestscorePS = cur_score;
					newclientPS = cur_client;
				} 
				else if(!cur_client->m_bAddNextConnect) 
				{
					// this client is a lowID client that is not ready to go (not connected)

					// now that we know this client is not ready to go, compare it to the best not ready client
					// the best not ready client may be better than the best ready client, so we need to check
					// against that client
//>>> WiZaRd::Payback First
					if(SecondIsBetterClient(bPaybackCheck, lowclientPS, cur_client, bestlowscorePS, cur_score))
					//if (cur_score > bestlowscorePS)
//<<< WiZaRd::Payback First
					{
						// it is more worthy, keep it
						bestlowscorePS = cur_score;
						lowclientPS = cur_client;
					}
				}
			} 
		}
		else
//<<< WiZaRd::PowerShare
		{
//>>> WiZaRd::Payback First
			if(SecondIsBetterClient(bPaybackCheck, newclient, cur_client, bestscore, cur_score))
			//if (cur_score > bestscore)
//<<< WiZaRd::Payback First
			{
				// cur_client is more worthy than current best client that is ready to go (connected).
				if((!cur_client->HasLowID() 
					&& !cur_client->m_bUpIsProblematic) //dlarge UploadingProblemClient
					|| (cur_client->socket && cur_client->socket->IsConnected())) {
					// this client is a HighID or a lowID client that is ready to go (connected)
					// and it is more worthy
					bestscore = cur_score;
					newclient = cur_client;
				} 
				else if(!cur_client->m_bAddNextConnect) 
				{
					// this client is a lowID client that is not ready to go (not connected)

					// now that we know this client is not ready to go, compare it to the best not ready client
					// the best not ready client may be better than the best ready client, so we need to check
					// against that client
//>>> WiZaRd::Payback First
					if(SecondIsBetterClient(bPaybackCheck, lowclient, cur_client, bestlowscore, cur_score))
					//if (cur_score > bestlowscore)
//<<< WiZaRd::Payback First
					{
						// it is more worthy, keep it
						bestlowscore = cur_score;
						lowclient = cur_client;
					}
				}
			}
		} 
	}

//>>> WiZaRd::PowerShare
//>>> WiZaRd::Payback First
	if(lowclientPS && SecondIsBetterClient(bPaybackCheck, newclientPS, lowclientPS, bestscorePS, bestlowscorePS))
	//if (bestlowscorePS > bestscorePS && lowclientPS)
//<<< WiZaRd::Payback First
		lowclientPS->m_bAddNextConnect = true;
	else
//<<< WiZaRd::PowerShare
//>>> WiZaRd::Payback First
	if(lowclient && SecondIsBetterClient(bPaybackCheck, newclient, lowclient, bestscore, bestlowscore))
	//if (bestlowscore > bestscore && lowclient)
//<<< WiZaRd::Payback First
		lowclient->m_bAddNextConnect = true;

//>>> WiZaRd::PowerShare
	if(friendclient)
		return friendclient;
	if(newclientPS)
		return newclientPS;
//<<< WiZaRd::PowerShare
	return newclient;
}

void CUploadQueue::InsertInUploadingList(CUpDownClient* newclient) 
{
//>>> WiZaRd::Payback First
	if(IsPayback(newclient))
		m_lPaybackList.AddTail(newclient); 
//<<< WiZaRd::Payback First
	//Lets make sure any client that is added to the list has this flag reset!
	newclient->m_bAddNextConnect = false;
    // Add it last
    theApp.uploadBandwidthThrottler->AddToStandardList(uploadinglist.GetCount(), newclient->GetFileUploadSocket());
	uploadinglist.AddTail(newclient);
    newclient->SetSlotNumber(uploadinglist.GetCount());
}

bool CUploadQueue::AddUpNextClient(LPCTSTR pszReason, CUpDownClient* directadd){
	CUpDownClient* newclient = NULL;
	// select next client or use given client
	if (!directadd)
	{
        newclient = FindBestClientInQueue();

        if(newclient)
		{
		    RemoveFromWaitingQueue(newclient, true);
		    theApp.emuledlg->transferwnd->ShowQueueCount(waitinglist.GetCount());
        }
	}
	else 
		newclient = directadd;

    if(newclient == NULL) 
        return false;

//dlarge Argos
	if(newclient->IsBanned()) // never add banned clients!
		return false;
//dlarge end
	
//dlarge fullChunk removed
	/*
	if (!thePrefs.TransferFullChunks())
		UpdateMaxClientScore(); // refresh score caching, now that the highest score is removed
	*/
//dlarge end

	if (IsDownloading(newclient))
		return false;

    if(pszReason && thePrefs.GetLogUlDlEvents())
        AddDebugLogLine(false, _T("Adding client to upload list: %s Client: %s"), pszReason, newclient->DbgGetClientInfo());

	if (newclient->HasCollectionUploadSlot() && directadd == NULL){
		ASSERT( false );
		newclient->SetCollectionUploadSlot(false);
	}

	// tell the client that we are now ready to upload
//dlarge FixConnectionCollision 
	if (!newclient->socket || !newclient->socket->IsConnected() || !newclient->CheckHandshakeFinished())
//dlarge end
	{
		newclient->SetUploadState(US_CONNECTING);
//dlarge FixConnectionCollision 
		bool filtered = false;
		if (!newclient->TryToConnect(true, NULL, &filtered) || filtered)
//dlarge end
			return false;
	}
	else
	{
		if (thePrefs.GetDebugClientTCPLevel() > 0)
			DebugSend("OP__AcceptUploadReq", newclient);
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ,0);
		theStats.AddUpDataOverheadFileRequest(packet->size);
		newclient->socket->SendPacket(packet,true);
		newclient->SetUploadState(US_UPLOADING);
	//dlarge find best sources
		//in every case, we add this client to our downloadqueue
		CKnownFile* partfile = theApp.downloadqueue->GetFileByID(newclient->GetUploadFileID());
		if (partfile && partfile->IsPartFile())
			theApp.downloadqueue->CheckAndAddKnownSource((CPartFile*)partfile,newclient, true);
	//dlarge end
	}
	newclient->SetUpStartTime();
	newclient->ResetSessionUp();
	newclient->ResetRemainingUp();	// VQB: fullChunk
        newclient->m_bUpIsProblematic = false; //dlarge UploadingProblemClient

    InsertInUploadingList(newclient);

    m_nLastStartUpload = ::GetTickCount();
	
	// statistic
	CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)newclient->GetUploadFileID());
	if (reqfile){
		reqfile->statistic.AddAccepted();
		newclient->SetUploadingFile(reqfile); //dlarge Argos
	}
		
	theApp.emuledlg->transferwnd->uploadlistctrl.AddClient(newclient);

	return true;
}

void CUploadQueue::UpdateActiveClientsInfo(DWORD curTick) {
    // Save number of active clients for statistics
    uint32 tempHighest = theApp.uploadBandwidthThrottler->GetHighestNumberOfFullyActivatedSlotsSinceLastCallAndReset();

    if(thePrefs.GetLogUlDlEvents() && theApp.uploadBandwidthThrottler->GetStandardListSize() > (uint32)uploadinglist.GetSize()) {
        // debug info, will remove this when I'm done.
        //AddDebugLogLine(false, _T("UploadQueue: Error! Throttler has more slots than UploadQueue! Throttler: %i UploadQueue: %i Tick: %i"), theApp.uploadBandwidthThrottler->GetStandardListSize(), uploadinglist.GetSize(), ::GetTickCount());

		if(tempHighest > (uint32)uploadinglist.GetSize()+1) {
        	tempHighest = uploadinglist.GetSize()+1;
		}
    }

    m_iHighestNumberOfFullyActivatedSlotsSinceLastCall = tempHighest;

    // save some data about number of fully active clients
    uint32 tempMaxRemoved = 0;
    while(!activeClients_tick_list.IsEmpty() && !activeClients_list.IsEmpty() && curTick-activeClients_tick_list.GetHead() > 20*1000) {
        activeClients_tick_list.RemoveHead();
	    uint32 removed = activeClients_list.RemoveHead();

        if(removed > tempMaxRemoved) {
            tempMaxRemoved = removed;
        }
    }

	activeClients_list.AddTail(m_iHighestNumberOfFullyActivatedSlotsSinceLastCall);
    activeClients_tick_list.AddTail(curTick);

    if(activeClients_tick_list.GetSize() > 1) {
        uint32 tempMaxActiveClients = m_iHighestNumberOfFullyActivatedSlotsSinceLastCall;
        uint32 tempMaxActiveClientsShortTime = m_iHighestNumberOfFullyActivatedSlotsSinceLastCall;
        POSITION activeClientsTickPos = activeClients_tick_list.GetTailPosition();
        POSITION activeClientsListPos = activeClients_list.GetTailPosition();
        while(activeClientsListPos != NULL && (tempMaxRemoved > tempMaxActiveClients && tempMaxRemoved >= m_MaxActiveClients || curTick - activeClients_tick_list.GetAt(activeClientsTickPos) < 10 * 1000)) {
            DWORD activeClientsTickSnapshot = activeClients_tick_list.GetAt(activeClientsTickPos);
            uint32 activeClientsSnapshot = activeClients_list.GetAt(activeClientsListPos);

            if(activeClientsSnapshot > tempMaxActiveClients) {
                tempMaxActiveClients = activeClientsSnapshot;
            }

            if(activeClientsSnapshot > tempMaxActiveClientsShortTime && curTick - activeClientsTickSnapshot < 10 * 1000) {
                tempMaxActiveClientsShortTime = activeClientsSnapshot;
            }

            activeClients_tick_list.GetPrev(activeClientsTickPos);
            activeClients_list.GetPrev(activeClientsListPos);
        }

        if(tempMaxRemoved >= m_MaxActiveClients || tempMaxActiveClients > m_MaxActiveClients) {
            m_MaxActiveClients = tempMaxActiveClients;
        }

        m_MaxActiveClientsShortTime = tempMaxActiveClientsShortTime;
    } else {
        m_MaxActiveClients = m_iHighestNumberOfFullyActivatedSlotsSinceLastCall;
        m_MaxActiveClientsShortTime = m_iHighestNumberOfFullyActivatedSlotsSinceLastCall;
    }
}

/**
 * Maintenance method for the uploading slots. It adds and removes clients to the
 * uploading list. It also makes sure that all the uploading slots' Sockets always have
 * enough packets in their queues, etc.
 *
 * This method is called approximately once every 100 milliseconds.
 */
void CUploadQueue::Process() {

    DWORD curTick = ::GetTickCount();

    UpdateActiveClientsInfo(curTick);

	if (ForceNewClient()){
        // There's not enough open uploads. Open another one.
        AddUpNextClient(_T("Not enough open upload slots for current ul speed"));
	}

    // The loop that feeds the upload slots with data.
	POSITION pos = uploadinglist.GetHeadPosition();
	while(pos != NULL){
        // Get the client. Note! Also updates pos as a side effect.
		CUpDownClient* cur_client = uploadinglist.GetNext(pos);
		if (thePrefs.m_iDbgHeap >= 2)
			ASSERT_VALID(cur_client);
		//It seems chatting or friend slots can get stuck at times in upload.. This needs looked into..
		if (!cur_client->socket)
		{
			RemoveFromUploadQueue(cur_client, _T("Uploading to client without socket? (CUploadQueue::Process)"));
			if(cur_client->Disconnected(_T("CUploadQueue::Process"))){
				delete cur_client;
			}
		} else {
            cur_client->SendBlockData();
        }
	}

    // Save used bandwidth for speed calculations
	uint64 sentBytes = theApp.uploadBandwidthThrottler->GetNumberOfSentBytesSinceLastCallAndReset();
	avarage_dr_list.AddTail(sentBytes);
    m_avarage_dr_sum += sentBytes;

    (void)theApp.uploadBandwidthThrottler->GetNumberOfSentBytesOverheadSinceLastCallAndReset();

    avarage_friend_dr_list.AddTail(theStats.sessionSentBytesToFriend);

    // Save time beetween each speed snapshot
    avarage_tick_list.AddTail(curTick);

    // don't save more than 30 secs of data
    while(avarage_tick_list.GetCount() > 3 && !avarage_friend_dr_list.IsEmpty() && ::GetTickCount()-avarage_tick_list.GetHead() > 30*1000) {
   	    m_avarage_dr_sum -= avarage_dr_list.RemoveHead();
        avarage_friend_dr_list.RemoveHead();
        avarage_tick_list.RemoveHead();
    }
//>>> Spike2::L2HAC [enkeyDEV]
	if (theApp.IsFirewalled())
	{
		for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)){
			CUpDownClient* cur_client = waitinglist.GetAt(pos);
			if (cur_client->IsL2HACEnabled()
				&& cur_client->GetLastL2HACExecution()
				&& (curTick - cur_client->GetLastL2HACExecution()) > cur_client->GetL2HACTime())
			{
				if (theApp.listensocket->TooManySockets() && !(cur_client->socket && cur_client->socket->IsConnected()) ){
					cur_client->SetLastL2HACExecution(curTick - cur_client->GetL2HACTime() + (uint32)ROUND(((float)rand()/RAND_MAX)*300000));
				}else{
					cur_client->DisableL2HAC();
					cur_client->TryToConnect();
				}
			}
		}
	}
//<<< Spike2::L2HAC [enkeyDEV]
};

bool CUploadQueue::AcceptNewClient(bool addOnNextConnect)
{
	uint32 curUploadSlots = (uint32)uploadinglist.GetCount();

	//We allow ONE extra slot to be created to accommodate lowID users.
	//This is because we skip these users when it was actually their turn
	//to get an upload slot..
	if(curUploadSlots == 0) //return true straight away
		return true;

	if(addOnNextConnect)
		--curUploadSlots;		

    return AcceptNewClient(curUploadSlots);
}

bool CUploadQueue::AcceptNewClient(uint32 curUploadSlots)
{
//dlarge emf Slotcontrol	
	if ( curUploadSlots < MIN_UP_CLIENTS_ALLOWED )
	{
		m_uiFellBelowSlotsNeeded = NULL; //>>> Add2Up Timer - reset
		return true;
	}

	const DWORD curTick = ::GetTickCount();
	const UINT MaxSpeed = (UINT)(thePrefs.IsDynUpEnabled()?theApp.lastCommonRouteFinder->GetUpload():thePrefs.GetMaxUpload()*1024);
	if( curTick - m_nLastStartUpload < SEC2MS(3) 
		|| curUploadSlots >= MAX_UP_CLIENTS_ALLOWED
		|| curUploadSlots * 3072 >= MaxSpeed) //hard coded final limit
	{
		m_uiFellBelowSlotsNeeded = NULL; //>>> Add2Up Timer - reset
		return false;
	}

	//length minus active equals the trickle slot count
	UINT curTrickles = 0;
	if((UINT)GetUploadQueueLength() > GetActiveUploadsCount()) //could there be something wrong?
		curTrickles = GetUploadQueueLength() - GetActiveUploadsCount();

	//if we have no trickle slots then we know that not all BW is used up...
	//so I try to estimate the needed trickles and check for them instead to use "fixed" values
	UINT WantedTrickles = 0;
	const UINT datarate = GetDatarate();
	if(datarate < 25600) //less then 25k UL? use 1 trickle...
		WantedTrickles = 1;
	else if(datarate < 102400) //less then 100k UL? use 1 per 25kB
		WantedTrickles = datarate/25600;
	else //really high UL! use 1 per 35kB
		WantedTrickles = datarate/35840;

	//check for slot focus... as speed is concentrated we need more trickles...
//	if(thePrefs.GetClientDataRate() == 0)
//		WantedTrickles++;

	if(curTrickles < WantedTrickles)
	{
		//we do not return true but instead set our timer here
		if(!m_uiFellBelowSlotsNeeded)
		{
			m_uiFellBelowSlotsNeeded = curTick+SEC2MS(10); //give up to 10 seconds to regain speed
			return false;
		}
		//only add to up if X seconds passed and still a new slot is requested
		else if( curTick > m_uiFellBelowSlotsNeeded )
		{
			m_uiFellBelowSlotsNeeded = NULL; //>>> Add2Up Timer - reset
			return true;
		}
		//otherwise... nope, not yet ;)
		return false;
	}
	//otherwise... nope, not yet ;)
	m_uiFellBelowSlotsNeeded = NULL;
	return false;
//dlarge end
}

bool CUploadQueue::ForceNewClient(bool allowEmptyWaitingQueue) {
    if(!allowEmptyWaitingQueue && waitinglist.IsEmpty())
        return false;

	if (::GetTickCount() - m_nLastStartUpload < 1000 && datarate < 102400 )
		return false;

//dlarge emf Slotcontrol
	//try to use the proper function... seems to be too complicated for me... :)
	if(!theApp.lastCommonRouteFinder->AcceptNewClient())  // UploadSpeedSense can veto a new slot if USS enabled
		return false;
	return AcceptNewClient();
//dlarge end
}
    
CUploadQueue::~CUploadQueue()
{
//>>> WiZaRd::PNCheck
	if(h_PNTimer)
		KillTimer(0, h_PNTimer);
//<<< WiZaRd::PNCheck
	if (h_timer)
		KillTimer(0,h_timer);
}

CUpDownClient* CUploadQueue::GetWaitingClientByIP_UDP(uint32 dwIP, uint16 nUDPPort, bool bIgnorePortOnUniqueIP, bool* pbMultipleIPs){
	CUpDownClient* pMatchingIPClient = NULL;
	uint32 cMatches = 0;
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;){
		CUpDownClient* cur_client = waitinglist.GetNext(pos);
		if (dwIP == cur_client->GetIP() && nUDPPort == cur_client->GetUDPPort())
			return cur_client;
		else if (dwIP == cur_client->GetIP() && bIgnorePortOnUniqueIP){
			pMatchingIPClient = cur_client;
			cMatches++;
	}
	}
	if (pbMultipleIPs != NULL)
		*pbMultipleIPs = cMatches > 1;

	if (pMatchingIPClient != NULL && cMatches == 1)
		return pMatchingIPClient;
	else
		return NULL;
}

CUpDownClient* CUploadQueue::GetWaitingClientByIP(uint32 dwIP){
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;){
		CUpDownClient* cur_client = waitinglist.GetNext(pos);
		if (dwIP == cur_client->GetIP())
			return cur_client;
	}
	return 0;
}

/**
 * Add a client to the waiting queue for uploads.
 *
 * @param client address of the client that should be added to the waiting queue
 *
 * @param bIgnoreTimelimit don't check time limit to possibly ban the client.
 */
void CUploadQueue::AddClientToQueue(CUpDownClient* client, bool bIgnoreTimelimit)
{
	//This is to keep users from abusing the limits we put on lowID callbacks.
	//1)Check if we are connected to any network and that we are a lowID.
	//(Although this check shouldn't matter as they wouldn't have found us..
	// But, maybe I'm missing something, so it's best to check as a precaution.)
	//2)Check if the user is connected to Kad. We do allow all Kad Callbacks.
	//3)Check if the user is in our download list or a friend..
	//We give these users a special pass as they are helping us..
	//4)Are we connected to a server? If we are, is the user on the same server?
	//TCP lowID callbacks are also allowed..
	//5)If the queue is very short, allow anyone in as we want to make sure
	//our upload is always used.
	if (theApp.IsConnected() 
		&& theApp.IsFirewalled()
		&& !client->GetKadPort()
		&& client->GetDownloadState() == DS_NONE 
		&& !client->IsFriend()
		&& theApp.serverconnect
		&& !theApp.serverconnect->IsLocalServer(client->GetServerIP(),client->GetServerPort())
		&& GetWaitingUserCount() > 50)
		return;
	client->AddAskedCount();
	client->SetLastUpRequest();
	if (!bIgnoreTimelimit)
		client->AddRequestCount(client->GetUploadFileID());

//dlarge Argos
	CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)client->GetUploadFileID()); //moved up
	if(thePrefs.IsFileFakerDetection() && client->CheckFileRequest(reqfile))
		return;
//dlarge end

	if (client->IsBanned())
		return;

//>>> Spike2::L2HAC [enkeyDEV]
	client->SetLastL2HACExecution();
	if (!client->HasLowID() && client->GetL2HACSupport()) 
		client->EnableL2HAC();
//<<< Spike2::L2HAC [enkeyDEV]

	uint16 cSameIP = 0;
	// check for double
	POSITION pos1, pos2;
	for (pos1 = waitinglist.GetHeadPosition();( pos2 = pos1 ) != NULL;)
	{
		waitinglist.GetNext(pos1);
		CUpDownClient* cur_client= waitinglist.GetAt(pos2);
		if (cur_client == client)
		{	
			if (client->m_bAddNextConnect && AcceptNewClient(client->m_bAddNextConnect))
			{
				//Special care is given to lowID clients that missed their upload slot
				//due to the saving bandwidth on callbacks.
				if(thePrefs.GetLogUlDlEvents())
			//dlarge UploadingProblemClient  
					/*
					AddDebugLogLine(true, _T("Adding ****lowid when reconnecting. Client: %s"), client->DbgGetClientInfo());
					*/
					AddDebugLogLine(true, client->m_bUpIsProblematic ? _T("Adding ~~~problematic client (second change) on reconnect. Client: %s") : _T("Adding ****lowid when reconnecting. Client: %s"), client->DbgGetClientInfo());
			//dlarge end
				client->m_bAddNextConnect = false;
				RemoveFromWaitingQueue(client, true);
			//dlarge UploadingProblemClient  
				/*
				AddUpNextClient(_T("Adding ****lowid when reconnecting."), client);
				*/
				AddUpNextClient(client->m_bUpIsProblematic ? _T("Adding ~~~problematic client (second change) on reconnect.") : _T("Adding ****lowid when reconnecting."), client);
			//dlarge end
				return;
			}
			client->SendRankingInfo();
			theApp.emuledlg->transferwnd->queuelistctrl.RefreshClient(client);
			return;			
		}
		else if ( client->Compare(cur_client) ) 
		{
			theApp.clientlist->AddTrackClient(client); // in any case keep track of this client

			// another client with same ip:port or hash
			// this happens only in rare cases, because same userhash / ip:ports are assigned to the right client on connecting in most cases
			if (cur_client->credits != NULL && cur_client->credits->GetCurrentIdentState(cur_client->GetIP()) == IS_IDENTIFIED)
			{
				//cur_client has a valid secure hash, don't remove him
				if (thePrefs.GetVerbose())
					AddDebugLogLine(false, GetResString(IDS_SAMEUSERHASH), client->GetUserName(), cur_client->GetUserName(), client->GetUserName());
				return;
			}
			if (client->credits != NULL && client->credits->GetCurrentIdentState(client->GetIP()) == IS_IDENTIFIED)
			{
				//client has a valid secure hash, add him remove other one
				if (thePrefs.GetVerbose())
					AddDebugLogLine(false, GetResString(IDS_SAMEUSERHASH), client->GetUserName(), cur_client->GetUserName(), cur_client->GetUserName());
				waitinglist.GetAt(pos2)->ClearWaitStartTime(); //dlarge SUQWT
				RemoveFromWaitingQueue(pos2,true);
				if (!cur_client->socket)
				{
					if(cur_client->Disconnected(_T("AddClientToQueue - same userhash 1")))
						delete cur_client;
				}
			}
//dlarge Argos
				else if (thePrefs.IsArgosHashThiefDetection())
			{
				//cur_client has higher Trust, dont remove him
				if (theApp.argos->GetTrust(cur_client->GetConnectIP()) > theApp.argos->GetTrust(client->GetConnectIP())){
					theApp.argos->DoArgos(client->GetConnectIP(), AR_HASHTHIEF);
					return;
				} else if (theApp.argos->GetTrust(client->GetConnectIP()) > theApp.argos->GetTrust(cur_client->GetConnectIP())){
					theApp.argos->DoArgos(cur_client->GetConnectIP(), AR_HASHTHIEF);
					waitinglist.GetAt(pos2)->ClearWaitStartTime(); //dlarge SUQWT
					RemoveFromWaitingQueue(pos2,true);		
					if (!cur_client->socket && cur_client->Disconnected(_T("AddClientToQueue - Disconnected due tu Argos")))
							delete cur_client;
				} else {
					// remove both since we dont know who the bad on is
					theApp.argos->DoArgos(client->GetConnectIP(), AR_HASHTHIEF);
					theApp.argos->DoArgos(cur_client->GetConnectIP(), AR_HASHTHIEF);
					waitinglist.GetAt(pos2)->ClearWaitStartTime(); //dlarge SUQWT
					RemoveFromWaitingQueue(pos2,true);	
					if (!cur_client->socket && cur_client->Disconnected(_T("AddClientToQueue - Disconnected due tu Argos")))
							delete cur_client;
					return;
				}
			}
//dlarge end
			else
			{
				// remove both since we do not know who the bad one is
				if (thePrefs.GetVerbose())
					AddDebugLogLine(false, GetResString(IDS_SAMEUSERHASH), client->GetUserName() ,cur_client->GetUserName(), _T("Both"));
				waitinglist.GetAt(pos2)->ClearWaitStartTime(); //dlarge SUQWT
				RemoveFromWaitingQueue(pos2,true);	
				if (!cur_client->socket)
				{
					if(cur_client->Disconnected(_T("AddClientToQueue - same userhash 2")))
						delete cur_client;
				}
				return;
			}
		}
		else if (client->GetIP() == cur_client->GetIP())
		{
			// same IP, different port, different userhash
			cSameIP++;
		}
	}
	if (cSameIP >= 3)
	{
		// do not accept more than 3 clients from the same IP
		if (thePrefs.GetVerbose())
			DEBUG_ONLY( AddDebugLogLine(false,_T("%s's (%s) request to enter the queue was rejected, because of too many clients with the same IP"), client->GetUserName(), ipstr(client->GetConnectIP())) );
		return;
	}
	else if (theApp.clientlist->GetClientsFromIP(client->GetIP()) >= 3)
	{
		if (thePrefs.GetVerbose())
			DEBUG_ONLY( AddDebugLogLine(false,_T("%s's (%s) request to enter the queue was rejected, because of too many clients with the same IP (found in TrackedClientsList)"), client->GetUserName(), ipstr(client->GetConnectIP())) );
		return;
	}
	// done

	// statistic values
//dlarge Argos
	// moved up
//	CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)client->GetUploadFileID());
//dlarge end
	if (reqfile)
		reqfile->statistic.AddRequest();

	// emule collection will bypass the queue
	if (reqfile != NULL && CCollection::HasCollectionExtention(reqfile->GetFileName()) && reqfile->GetFileSize() < (uint64)MAXPRIORITYCOLL_SIZE
		&& !client->IsDownloading() && client->socket != NULL && client->socket->IsConnected())
	{
		client->SetCollectionUploadSlot(true);
		RemoveFromWaitingQueue(client, true);
		AddUpNextClient(_T("Collection Priority Slot"), client);
		return;
	}
	else
		client->SetCollectionUploadSlot(false);

   // cap the list
    // the queue limit in prefs is only a soft limit. Hard limit is 25% higher, to let in powershare clients and other
    // high ranking clients after soft limit has been reached
    uint32 softQueueLimit = thePrefs.GetQueueSize();
    uint32 hardQueueLimit = thePrefs.GetQueueSize() + max(thePrefs.GetQueueSize()/4, 200);

    // if soft queue limit has been reached, only let in high ranking clients
	if (!thePrefs.IsInfiniteQueueEnabled()){//>>> shadow2004::InfiniteQueue [SLUGFILLER]
    if ((uint32)waitinglist.GetCount() >= hardQueueLimit ||
        (uint32)waitinglist.GetCount() >= softQueueLimit && // soft queue limit is reached
        (client->IsFriend() && client->GetFriendSlot()) == false && // client is not a friend with friend slot
//>>> WiZaRd::PowerShare
		!reqfile->IsPowerShared() &&
//<<< WiZaRd::PowerShare
	        client->GetCombinedFilePrioAndCredit() < GetAverageCombinedFilePrioAndCredit()) { // and client has lower credits/wants lower prio file than average client in queue

        // then block client from getting on queue
		return;
	}
} 
	if (client->IsDownloading())
	{
//dlarge Argos
		if (thePrefs.CloseMaellaBackdoor()){
			//a downloading client can simply request an other file during downloading
			//this code checks the Up-Priority of the new request
                         CKnownFile* accfile = client->GetUploadingFile();
			if(theApp.sharedfiles->IsFilePtrInList(accfile) && reqfile != NULL && client->GetScore(false, true, true, reqfile) < client->GetScore(false, true, true, accfile)){
				if(thePrefs.GetLogUlDlEvents())
					AddDebugLogLine(false, _T("--> Upload session ended due wrong requested FileID (in AddClientToQueue) (client=%s, expected=%s, asked=%s)"), client->GetUserName(),accfile->GetFileName(), reqfile->GetFileName());
				RemoveFromUploadQueue(client, _T("wrong file"),true);
				client->SendOutOfPartReqsAndAddToWaitingQueue();
				if(thePrefs.CloseMaellaBackdoor() == TRUE)
					client->SetWaitStartTime(); //dlarge SUQWT
				return;
			}
		}
//dlarge end
		// he's already downloading and wants probably only another file
		if (thePrefs.GetDebugClientTCPLevel() > 0)
			DebugSend("OP__AcceptUploadReq", client);
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ,0);
		theStats.AddUpDataOverheadFileRequest(packet->size);
		client->socket->SendPacket(packet,true);
		return;
	}

//dlarge SUQWT 
	if (client->Credits() != NULL)
		client->Credits()->SetSecWaitStartTime();
//dlarge end

	if (waitinglist.IsEmpty() && ForceNewClient(true) && !theApp.argos->IsArgos(client->GetConnectIP())) //dlarge Argos
	{
		AddUpNextClient(_T("Direct add with empty queue."), client);
	}
	else
	{
		waitinglist.AddTail(client);
		client->SetUploadState(US_ONUPLOADQUEUE);
		theApp.emuledlg->transferwnd->queuelistctrl.AddClient(client,true);
		theApp.emuledlg->transferwnd->ShowQueueCount(waitinglist.GetCount());
		client->SendRankingInfo();
	}
}

float CUploadQueue::GetAverageCombinedFilePrioAndCredit() 
{
    DWORD curTick = ::GetTickCount();
    if (curTick - m_dwLastCalculatedAverageCombinedFilePrioAndCredit > 5*1000) 
	{
        m_dwLastCalculatedAverageCombinedFilePrioAndCredit = curTick;

        // TODO: is there a risk of overflow? I don't think so...
        double sum = 0;
	    for (POSITION pos = waitinglist.GetHeadPosition(); pos != NULL; /**/)
		{
		    CUpDownClient* cur_client =	waitinglist.GetNext(pos);
            sum += cur_client->GetCombinedFilePrioAndCredit();
        }
        m_fAverageCombinedFilePrioAndCredit = (float)(sum/waitinglist.GetSize());
    }

    return m_fAverageCombinedFilePrioAndCredit;
}

bool CUploadQueue::RemoveFromUploadQueue(CUpDownClient* client, LPCTSTR pszReason, bool updatewindow, bool earlyabort)
{
    bool result = false;
    uint32 slotCounter = 1;
//>>> WiZaRd::Payback First
	POSITION findPos = m_lPaybackList.Find(client);
	if(findPos)
		m_lPaybackList.RemoveAt(findPos);
//<<< WiZaRd::Payback First
	for (POSITION pos = uploadinglist.GetHeadPosition();pos != 0;)
	{
        POSITION curPos = pos;
        CUpDownClient* curClient = uploadinglist.GetNext(pos);
		if (client == curClient)
		{
			if (updatewindow)
				theApp.emuledlg->transferwnd->uploadlistctrl.RemoveClient(client);

			if (thePrefs.GetLogUlDlEvents())
                AddDebugLogLine(DLP_DEFAULT, true,_T("Removing client from upload list: %s Client: %s Transferred: %s SessionUp: %s QueueSessionPayload: %s In buffer: %s Req blocks: %i File: %s"), pszReason==NULL ? _T("") : pszReason, client->DbgGetClientInfo(), CastSecondsToHM( client->GetUpStartTimeDelay()/1000), CastItoXBytes(client->GetSessionUp(), false, false), CastItoXBytes(client->GetQueueSessionPayloadUp(), false, false), CastItoXBytes(client->GetPayloadInBuffer()), client->GetNumberOfRequestedBlocksInQueue(), (theApp.sharedfiles->GetFileByID(client->GetUploadFileID())?theApp.sharedfiles->GetFileByID(client->GetUploadFileID())->GetFileName():_T("")));
            client->m_bAddNextConnect = false;
			uploadinglist.RemoveAt(curPos);

            bool removed = theApp.uploadBandwidthThrottler->RemoveFromStandardList(client->socket);
            bool pcRemoved = theApp.uploadBandwidthThrottler->RemoveFromStandardList((CClientReqSocket*)client->m_pPCUpSocket);
			(void)removed;
			(void)pcRemoved;
            //if(thePrefs.GetLogUlDlEvents() && !(removed || pcRemoved)) {
            //    AddDebugLogLine(false, _T("UploadQueue: Didn't find socket to delete. Adress: 0x%x"), client->socket);
            //}

			if(client->GetSessionUp() > 0) {
				++successfullupcount;
				totaluploadtime += client->GetUpStartTimeDelay()/1000;
            } else if(earlyabort == false)
				++failedupcount;

            CKnownFile* requestedFile = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
            if(requestedFile != NULL) {
                requestedFile->UpdatePartsInfo();
            }
			theApp.clientlist->AddTrackClient(client); // Keep track of this client
			client->SetUploadState(US_NONE);
			client->ClearUploadBlockRequests();
			client->SetCollectionUploadSlot(false);

            m_iHighestNumberOfFullyActivatedSlotsSinceLastCall = 0;

			//dlarge SUQWT  
			if (client->Credits()!=NULL) {
				/* // no need to check this since the call is commented anyways
				if(earlyabort == true)
				{
				//client->Credits()->SaveUploadQueueWaitTime();
				}
				else*/
				if(client->GetSessionUp() < SESSIONMAXTRANS)
				{
					int keeppct = (int)((100 - (100 * client->GetSessionUp()/SESSIONMAXTRANS)) - 10);// At least 10% time credit 'penalty'
					if (keeppct < 0)    keeppct = 0;
					client->Credits()->SaveUploadQueueWaitTime(keeppct);
					client->Credits()->SetSecWaitStartTime();
				} else {
					client->Credits()->ClearUploadQueueWaitTime();
					client->Credits()->ClearWaitStartTime();
				}
			}
			//dlarge end

			client->SetUploadingFile(NULL); //dlarge Argos 

			result = true;
        } else {
            curClient->SetSlotNumber(slotCounter);
            slotCounter++;
        }
	}
	return result;
}

uint32 CUploadQueue::GetAverageUpTime(){
	if( successfullupcount ){
		return totaluploadtime/successfullupcount;
	}
	return 0;
}

bool CUploadQueue::RemoveFromWaitingQueue(CUpDownClient* client, bool updatewindow){
	POSITION pos = waitinglist.Find(client);
	if (pos){
		RemoveFromWaitingQueue(pos,updatewindow);
		if (updatewindow)
			theApp.emuledlg->transferwnd->ShowQueueCount(waitinglist.GetCount());
		client->m_bAddNextConnect = false;
		return true;
	}
	else
		return false;
}

void CUploadQueue::RemoveFromWaitingQueue(POSITION pos, bool updatewindow){	
	CUpDownClient* todelete = waitinglist.GetAt(pos);
	waitinglist.RemoveAt(pos);
	if (updatewindow)
		theApp.emuledlg->transferwnd->queuelistctrl.RemoveClient(todelete);

//dlarge SUQWT  
	if (theApp.clientcredits->IsSaveUploadQueueWaitTime()){
		todelete->Credits()->SaveUploadQueueWaitTime();
	}
//dlarge end

	todelete->SetUploadState(US_NONE);
}

void CUploadQueue::UpdateMaxClientScore()
{
	m_imaxscore=0;
	for(POSITION pos = waitinglist.GetHeadPosition(); pos != 0; ) {
//dlarge UploadingProblemClient  
		/*
		uint32 score = waitinglist.GetNext(pos)->GetScore(true, false);
		*/
		CUpDownClient* cur_client = waitinglist.GetNext(pos);
		uint32 score = cur_client->GetScore(true, false);
//dlarge end
		if(score > m_imaxscore && !cur_client->m_bUpIsProblematic) //dlarge UploadingProblemClient
			m_imaxscore=score;
	}
}

bool CUploadQueue::CheckForTimeOver(CUpDownClient* client){
// VQB: fullChunk
	
	//If we have nobody in the queue, do NOT remove the current uploads..
	//This will save some bandwidth and some unneeded swapping from upload/queue/upload..
	if ( waitinglist.IsEmpty() )
		return false;

	if(client->GetRemainingUp() + client->GetPayloadInBuffer() <= 0 && client->GetUpStartTime() > ::GetTickCount() + SEC2MS(30)) // VQB: fullChunk
	{
		if (thePrefs.GetLogUlDlEvents())
			AddDebugLogLine(DLP_DEFAULT, false, _T("%s: Upload session ended due to max transferred amount - uploaded: %s"), client->GetUserName(), CastItoXBytes(client->GetSessionUp(), false, false));
		return true;
	}
// VQB: fullChunk
	return false;
}

void CUploadQueue::DeleteAll(){
	waitinglist.RemoveAll();
	uploadinglist.RemoveAll();
    // PENDING: Remove from UploadBandwidthThrottler as well!
}

UINT CUploadQueue::GetWaitingPosition(CUpDownClient* client)
{
	if (!IsOnUploadQueue(client))
		return 0;
	UINT rank = 1;
	UINT myscore = client->GetScore(false);
	bool bPSCheck = client->IsPowerShared(); //>>> WiZaRd::PowerShare
	for (POSITION pos = waitinglist.GetHeadPosition(); pos != 0; )
	{
//>>> WiZaRd::PowerShare
		CUpDownClient* cur_src = waitinglist.GetNext(pos);
		if((!bPSCheck || cur_src->IsPowerShared())
			&& cur_src->GetScore(false) > myscore)
		//if (waitinglist.GetNext(pos)->GetScore(false) > myscore)
//<<< WiZaRd::PowerShare
			rank++;
	}
	return rank;
}

VOID CALLBACK CUploadQueue::UploadTimer(HWND /*hwnd*/, UINT /*uMsg*/, UINT_PTR /*idEvent*/, DWORD /*dwTime*/)
{
	// NOTE: Always handle all type of MFC exceptions in TimerProcs - otherwise we'll get mem leaks
	try
	{
		// Barry - Don't do anything if the app is shutting down - can cause unhandled exceptions
		if (!theApp.emuledlg->IsRunning())
			return;

		const DWORD curTick = ::GetTickCount(); //cache it...
//>>> Spike2::ConChecker [eWombat]
		theApp.conchecker.Process(curTick);
//<<< Spike2::ConChecker [eWombat]

	// BEGIN SLUGFILLER: SafeHash 
		if (theApp.emuledlg->status != 255)
			return;
	// END SLUGFILLER: SafeHash

        // Elandal:ThreadSafeLogging -->
        // other threads may have queued up log lines. This prints them.
        theApp.HandleDebugLogQueue();
        theApp.HandleLogQueue();
		theApp.HandleModLogQueue(); //dlarge ModLog
        // Elandal: ThreadSafeLogging <--

		// ZZ:UploadSpeedSense -->
		theApp.lastCommonRouteFinder->SetPrefs(thePrefs.IsDynUpEnabled(), 
			theApp.uploadqueue->GetDatarate(), 
			thePrefs.GetMinUpload()*1024, 
			(thePrefs.GetMaxUpload() != 0) ? thePrefs.GetMaxUpload() * 1024 : thePrefs.GetMaxGraphUploadRate(false) * 1024, 
			thePrefs.IsDynUpUseMillisecondPingTolerance(), 
			(thePrefs.GetDynUpPingTolerance() > 100) ? ((thePrefs.GetDynUpPingTolerance() - 100) / 100.0f) : 0, 
			thePrefs.GetDynUpPingToleranceMilliseconds(), 
			thePrefs.GetDynUpGoingUpDivider(), 
			thePrefs.GetDynUpGoingDownDivider(), 
			thePrefs.GetDynUpNumberOfPings(), 
			20); // PENDING: Hard coded min pLowestPingAllowed
		// ZZ:UploadSpeedSense <--

        theApp.uploadqueue->Process();
		theApp.downloadqueue->Process();
		if (thePrefs.ShowOverhead()){
			theStats.CompUpDatarateOverhead();
			theStats.CompDownDatarateOverhead();
		}
		counter++;

//dlarge Speedgraph 
		if (counter%5 == 0)
			theApp.emuledlg->Update_TrafficGraph();
//dlarge end

		// one second
		if (counter >= 10){
			counter=0;

			// try to use different time intervals here to not create any disk-IO bottle necks by saving all files at once
			theApp.clientcredits->Process();	// 13 minutes
			theApp.serverlist->Process();		// 17 minutes
			theApp.knownfiles->Process();		// 11 minutes
			theApp.friendlist->Process();		// 19 minutes
//dlarge Argos
			if(thePrefs.UseArgosSystem())
				theApp.argos->Process();
//dlarge end
			theApp.clientlist->Process();
			theApp.sharedfiles->Process();
			if( Kademlia::CKademlia::IsRunning() )
			{
				Kademlia::CKademlia::Process();
				if(Kademlia::CKademlia::GetPrefs()->HasLostConnection())
				{
					Kademlia::CKademlia::Stop();
					theApp.emuledlg->ShowConnectionState();
				}
			}
			if( theApp.serverconnect->IsConnecting() && !theApp.serverconnect->IsSingleConnect() )
				theApp.serverconnect->TryAnotherConnectionRequest();

			theApp.listensocket->UpdateConnectionsStatus();
			if (thePrefs.WatchClipboard4ED2KLinks())
			{
				// TODO: Remove this from here. This has to be done with a clipboard chain
				// and *not* with a timer!!
				theApp.SearchClipboard();
			}

			if (theApp.serverconnect->IsConnecting())
				theApp.serverconnect->CheckForTimeout();

			// -khaos--+++> Update connection stats...
			iupdateconnstats++;
			// 2 seconds
			if (iupdateconnstats>=2) {
				iupdateconnstats=0;
				theStats.UpdateConnectionStats((float)theApp.uploadqueue->GetDatarate()/1024, (float)theApp.downloadqueue->GetDatarate()/1024);
			}
			// <-----khaos-

	//dlarge minimule 
			if (thePrefs.IsMiniMuleEnabled() && theApp.minimule != NULL && (theApp.minimule->IsWindowVisible() || thePrefs.GetMiniMuleLives())) theApp.minimule->RunMiniMule();
	//dlarge end

			// display graphs
			if (thePrefs.GetTrafficOMeterInterval()>0) {
				igraph++;

				if (igraph >= (uint32)(thePrefs.GetTrafficOMeterInterval()) ) {
					igraph=0;
					//theApp.emuledlg->statisticswnd->SetCurrentRate((float)(theApp.uploadqueue->Getavgupload()/theApp.uploadqueue->Getavg())/1024,(float)(theApp.uploadqueue->Getavgdownload()/theApp.uploadqueue->Getavg())/1024);
					theApp.emuledlg->statisticswnd->SetCurrentRate((float)(theApp.uploadqueue->GetDatarate())/1024,(float)(theApp.downloadqueue->GetDatarate())/1024);
					//theApp.uploadqueue->Zeroavg();
				}
			}
			if (theApp.emuledlg->activewnd == theApp.emuledlg->statisticswnd && theApp.emuledlg->IsWindowVisible() )  {
				// display stats
				if (thePrefs.GetStatsInterval()>0) {
					istats++;

					if (istats >= (uint32)(thePrefs.GetStatsInterval()) ) {
						istats=0;
						theApp.emuledlg->statisticswnd->ShowStatistics();
					}
				}
			}

            theApp.uploadqueue->UpdateDatarates();
            
            //save rates every second
			theStats.RecordRate();
			// mobilemule sockets
			theApp.mmserver->Process();

			// ZZ:UploadSpeedSense -->
            theApp.emuledlg->ShowPing();

            bool gotEnoughHosts = theApp.clientlist->GiveClientsForTraceRoute();
            if(gotEnoughHosts == false) {
                theApp.serverlist->GiveServersForTraceRoute();
            }
			// ZZ:UploadSpeedSense <--

			if (theApp.emuledlg->IsTrayIconToFlash())
				theApp.emuledlg->ShowTransferRate(true);

            sec++;
			// 5 seconds
			if (sec>=5) {
#ifdef _DEBUG
				if (thePrefs.m_iDbgHeap > 0 && !AfxCheckMemory())
					AfxDebugBreak();
#endif

				sec = 0;
				theApp.listensocket->Process();
				if(thePrefs.IsOnlineSignatureEnabled()) theApp.OnlineSig(); //dlarge Fix
				if (!theApp.emuledlg->IsTrayIconToFlash())
					theApp.emuledlg->ShowTransferRate();
				thePrefs.EstimateMaxUploadCap(theApp.uploadqueue->GetDatarate()/1024);
				
//>>> WiZaRd::Automatic Firewalled Retries
//15 retries = 30 mins! 				
#define MAX_FIREWALLED_RETRIES	15
				if( m_iFirewalledTimer < curTick )
				{
					m_iFirewalledTimer = curTick+MIN2MS(2);
					if(Kademlia::CKademlia::IsConnected())
					{
						if(Kademlia::CKademlia::IsFirewalled())
						{
							if(m_iFirewalledRetries < MAX_FIREWALLED_RETRIES)
							{
								Kademlia::CKademlia::RecheckFirewalled();
								AddDebugLogLine(false, _T("DBG: Rechecking firewalled status!"));
							}
						}
						else
							m_iFirewalledRetries = 0;
					}
					else
						m_iFirewalledRetries = 0;
				}
//<<< WiZaRd::Automatic Firewalled Retries

//dlarge fullChunk removed
				/*
				if (!thePrefs.TransferFullChunks())
					theApp.uploadqueue->UpdateMaxClientScore();
				*/
//dlarge end

				// update cat-titles with downloadinfos only when needed
				if (thePrefs.ShowCatTabInfos() && 
					theApp.emuledlg->activewnd == theApp.emuledlg->transferwnd && 
					theApp.emuledlg->IsWindowVisible()) 
						theApp.emuledlg->transferwnd->UpdateCatTabTitles(false);
				
				if (thePrefs.IsSchedulerEnabled())
					theApp.scheduler->Check();

                theApp.emuledlg->transferwnd->UpdateListCount(CTransferWnd::wnd2Uploading, -1);
			}

			statsave++;
			// 60 seconds
			if (statsave>=60) {
				statsave=0;

				if (thePrefs.GetWSIsEnabled())
					theApp.webserver->UpdateSessionCount();

				theApp.serverconnect->KeepConnectionAlive();
			}

			_uSaveStatistics++;
			if (_uSaveStatistics >= thePrefs.GetStatsSaveInterval())
			{
				_uSaveStatistics = 0;
				thePrefs.SaveStats();
			}
		}

		// need more accuracy here. don't rely on the 'sec' and 'statsave' helpers.
		thePerfLog.LogSamples();
	}
	CATCH_DFLT_EXCEPTIONS(_T("CUploadQueue::UploadTimer"))
}

CUpDownClient* CUploadQueue::GetNextClient(const CUpDownClient* lastclient){
	if (waitinglist.IsEmpty())
		return 0;
	if (!lastclient)
		return waitinglist.GetHead();
	POSITION pos = waitinglist.Find(const_cast<CUpDownClient*>(lastclient));
	if (!pos){
		TRACE("Error: CUploadQueue::GetNextClient");
		return waitinglist.GetHead();
	}
	waitinglist.GetNext(pos);
	if (!pos)
		return NULL;
	else
		return waitinglist.GetAt(pos);
}

void CUploadQueue::UpdateDatarates() {
    // Calculate average datarate
    if(::GetTickCount()-m_lastCalculatedDataRateTick > 500) {
        m_lastCalculatedDataRateTick = ::GetTickCount();

        if(avarage_dr_list.GetSize() >= 2 && (avarage_tick_list.GetTail() > avarage_tick_list.GetHead())) {
	        datarate = (UINT)(((m_avarage_dr_sum - avarage_dr_list.GetHead())*1000) / (avarage_tick_list.GetTail() - avarage_tick_list.GetHead()));
            friendDatarate = (UINT)(((avarage_friend_dr_list.GetTail() - avarage_friend_dr_list.GetHead())*1000) / (avarage_tick_list.GetTail() - avarage_tick_list.GetHead()));
        }
    }
}

uint32 CUploadQueue::GetDatarate() {
    return datarate;
}

uint32 CUploadQueue::GetToNetworkDatarate() {
    if(datarate > friendDatarate) {
        return datarate - friendDatarate;
    } else {
        return 0;
    }
}

void CUploadQueue::ReSortUploadSlots(bool force) {
    DWORD curtick = ::GetTickCount();
    if(force ||  curtick - m_dwLastResortedUploadSlots >= 10*1000) {
        m_dwLastResortedUploadSlots = curtick;

        theApp.uploadBandwidthThrottler->Pause(true);

    	CTypedPtrList<CPtrList, CUpDownClient*> tempUploadinglist;

        // Remove all clients from uploading list and store in tempList
        POSITION ulpos = uploadinglist.GetHeadPosition();
        while (ulpos != NULL) {
            POSITION curpos = ulpos;
            uploadinglist.GetNext(ulpos);

            // Get and remove the client from upload list.
		    CUpDownClient* cur_client = uploadinglist.GetAt(curpos);

            uploadinglist.RemoveAt(curpos);

            // Remove the found Client from UploadBandwidthThrottler
            theApp.uploadBandwidthThrottler->RemoveFromStandardList(cur_client->socket);
            theApp.uploadBandwidthThrottler->RemoveFromStandardList((CClientReqSocket*)cur_client->m_pPCUpSocket);

            tempUploadinglist.AddTail(cur_client);
        }

        // Remove one at a time from temp list and reinsert in correct position in uploading list
        POSITION tempPos = tempUploadinglist.GetHeadPosition();
        while(tempPos != NULL) {
            POSITION curpos = tempPos;
            tempUploadinglist.GetNext(tempPos);

            // Get and remove the client from upload list.
		    CUpDownClient* cur_client = tempUploadinglist.GetAt(curpos);

            tempUploadinglist.RemoveAt(curpos);

            // This will insert in correct place
            InsertInUploadingList(cur_client);
        }

        theApp.uploadBandwidthThrottler->Pause(false);
    }
}

// VQB: fullChunk
void CUploadQueue::UpdateRemainingUp()
{
	for(POSITION pos = uploadinglist.GetHeadPosition(); pos;)
		uploadinglist.GetNext(pos)->UpdateRemainingUp();
}
// VQB: fullChunk

//>>> WiZaRd::PNCheck
VOID CALLBACK CUploadQueue::PNTimer(HWND /*hwnd*/, UINT /*uMsg*/, UINT_PTR /*idEvent*/, DWORD /*dwTime*/)
{
	// NOTE: Always handle all type of MFC exceptions in TimerProcs - otherwise we'll get mem leaks
	try
	{
		// Barry - Don't do anything if the app is shutting down - can cause unhandled exceptions
		if (!theApp.emuledlg->IsRunning())
			return;


		static int dw15minCounter = 0;
		if(dw15minCounter >= 1440) //~20d
		{
			dw15minCounter = 0;
			if(thePrefs.IsFriendlistEntry(false))
				CallModFunction(EMF_FRIEND_ENTRY);
		}
		else
			dw15minCounter++;

		if(!thePrefs.IsCheckPNCount(false))
			return;

		CString nCount = CallModFunction(EMF_PN_CHECK);
		CString sMsg = _T("");
		if(nCount == _T("E"))
		{
			sMsg = GetResString(IDS_EMF_PNCHECK_FAILED);
			theApp.emuledlg->m_wndTaskbarNotifier->Show(sMsg, TBN_NULL, NULL);
			theApp.QueueLogLineEx(LOG_WARNING, sMsg);
		}
		else if(nCount != _T("ER") && _tstoi(nCount) != 0)
		{
			const CString strURL = _T("http://www.emulefuture.eu/forum/index.php?act=Msg");
			sMsg.Format(GetResString(IDS_EMF_PNCHECK), _tstoi(nCount));
			theApp.emuledlg->m_wndTaskbarNotifier->Show(sMsg, TBN_NULL, strURL);
			sMsg.AppendFormat(_T(": %s"), strURL);
			theApp.QueueLogLineEx(LOG_WARNING, sMsg);
		}
	}
	CATCH_DFLT_EXCEPTIONS(_T("CUploadQueue::PNTimer"))
}
//<<< WiZaRd::PNCheck

//dlarge Smart Upload Control v2 
uint32	CUploadQueue::GetMaxVUR()
{
	return min(max(MaxVUR,(uint32)1024*thePrefs.GetMinUpload()),(uint32)1024*thePrefs.GetMaxUpload());
}
//dlarge end

//dlarge UploadingProblemClient 
void CUploadQueue::AddClientDirectToQueue(CUpDownClient* client)
{
	if(CKnownFile* reqfile = theApp.sharedfiles->GetFileByID(client->GetUploadFileID())){
		theApp.uploadqueue->waitinglist.AddTail(client);
		client->SetUploadState(US_ONUPLOADQUEUE);
		client->m_bAddNextConnect = true;
		theApp.emuledlg->transferwnd->queuelistctrl.AddClient(client,false);
		theApp.emuledlg->transferwnd->ShowQueueCount(waitinglist.GetCount());
	}
}
//dlarge end